import 'dart:async';

abstract class State {
  Map<String, dynamic> toJson();
}

class Store {
  Map<Type, State> _current;
  final StreamController<Store> _changeController;

  Store(List<State> stores, {
    bool syncStream = true,
  })  : _changeController = StreamController.broadcast(sync: syncStream) {
    _current = Map.unmodifiable({ for (var store in stores) store.runtimeType : store });
  }

  Map<Type, State> get current => _current;

  T getState<T extends State>() => _current[T];

  Store unsafeSetState<T extends State>(T value) {
    final newState = {..._current};
    newState[T] = value;
    _current = Map.unmodifiable(newState);
    return this;
  }

  Store setState<T extends State>(T value) =>
      unsafeSetState<T>(value).forceEmit();

  Store updateState<T extends State>(T Function(T) updateFn) =>
      setState<T>(updateFn(getState<T>()));

  Store unsafeUpdateState<T extends State>(T Function(T) updateFn) =>
      unsafeSetState<T>(updateFn(getState<T>()));

  Stream<Store> get onChange => _changeController.stream;

  Map<String, dynamic> toJson() => _current.map((key, value) => MapEntry(key.toString(), value.toJson()));

  Future teardown() {
    _current = Map.identity();
    return _changeController.close();
  }

  Store forceEmit() {
    _changeController.add(this);
    return this;
  }
}